﻿using System;
using System.Collections.Concurrent;
using System.Diagnostics;
using System.Threading;

namespace Platform.Events
{
    public class EventQueue<TEventArgs> : IDisposable
        where TEventArgs : EventArgs
    {
        private IEventSubscription<TEventArgs> _Subscriber;
        private BlockingCollection<TEventArgs> _Collection;
        private ConcurrentQueue<TEventArgs> _Queue;

        public EventQueue( IEventSubscription<TEventArgs> subscriber, int size = 2048)
        {
            _Subscriber = subscriber;
            _Subscriber.Raised += OnEventRaised;

            _Queue = new ConcurrentQueue<TEventArgs>();
            _Collection = new BlockingCollection<TEventArgs>( _Queue, size );
        }

        #region IDisposable Support

        private bool disposedValue = false; // To detect redundant calls

        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            Dispose(true);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    // close subscriber
                    _Subscriber.Dispose();
                    _Subscriber = null;

                    // close cancellation source
                    _Collection.CompleteAdding();
                    _Collection.Dispose();
                    _Collection = null;
                }

                disposedValue = true;
            }
        }

        #endregion

        private void OnEventRaised(object sender, TEventArgs e)
        {
            try
            {
                _Collection.TryAdd(e, 5000);
            }
            catch( Exception ex )
            {
                Debug.WriteLine($"EventQueue failed to add item: {ex.Message}");
            }
        }

        public bool TryTake( out TEventArgs args, TimeSpan timeout, CancellationToken token )
        {
            return _Collection.TryTake(out args, (int)timeout.TotalMilliseconds, token);
        }
    }
}
