﻿using System;
using System.IO;
using System.Threading;

#if MF_FRAMEWORK
using Microsoft.SPOT;
#endif

namespace Platform
{
    /// <summary>
    /// Generic Excption Handler
    /// </summary>
    public static class ExceptionHandler
    {
        #region Single Place Exception Handling

        /// <summary>
        /// Signals an exception has occured
        /// </summary>
        public static event ErrorEventHandler Exception;

        /// <summary>
        /// Single Place exception handling
        /// </summary>
        /// <param name="source"></param>
        /// <param name="exception"></param>
        public static void Catch(object source, Exception exception)
        {
            var handler = Exception;
            if (handler != null) handler(source, new ErrorEventArgs( exception ) );
        }

        /// <summary>
        /// Lock and catch exceptions
        /// </summary>
        /// <param name="source"></param>
        /// <param name="action"></param>
        public static bool LockAndCatch(object source, Action action)
        {
            lock (source)
            {
                try
                {
                    action();
                    return true;
                }
                catch (Exception ex)
                {
                    Catch(source, ex);
                    return false;
                }
            }
        }

        /// <summary>
        /// Process action and catch uncought exceptions
        /// </summary>
        /// <param name="source"></param>
        /// <param name="action"></param>
        public static bool Catch(object source, Action action)
        {
            try
            {
                action();
                return true;
            }
            catch (Exception ex)
            {
                Catch(source, ex);
                return false;
            }
        }

        /// <summary>
        /// Signals an exception has occured
        /// </summary>
        public static event ErrorEventHandler Ignored;

        /// <summary>
        /// Single Place exception handling
        /// </summary>
        /// <param name="source"></param>
        /// <param name="exception"></param>
        public static void Ignore(object source, Exception ex)
        {
            var handler = Ignored;
            if (handler != null) handler(source, new ErrorEventArgs(ex));
        }

        /// <summary>
        /// Run this action and ignore any exceptions
        /// </summary>
        /// <param name="action"></param>
        public static void Ignore(object source, Action action)
        {
            try
            {
                action();
            }
            catch (Exception ex)
            {
                Ignore(source, ex);
            }
        }

        #endregion

        #region retry actions on exception

        /// <summary>
        /// Retry and throw after all tries have been done
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="count"></param>
        /// <param name="sleep"></param>
        /// <param name="func"></param>
        /// <returns></returns>
        public static T Retry<T>(int count, int sleep, Func<T> func)
        {
            while (true)
            {
                try
                {
                    return func();
                }
                catch (Exception)
                {
                    if (--count == 0) throw;
                    else if (sleep > 0) Thread.Sleep(sleep);
                }
            }
        }

        /// <summary>
        /// Retry and throw after all tries have been done
        /// </summary>
        /// <param name="count"></param>
        /// <param name="action"></param>
        public static void Retry(int count, int sleep, Action action)
        {
            while (true)
            {
                try
                {
                    action();
                    return;
                }
                catch (Exception)
                {
                    if (--count == 0) throw;
                    else if (sleep > 0) Thread.Sleep(sleep);
                }
            }
        }

        #endregion
    }
}
