﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.NetworkInformation;
using System.Globalization;
using System.Text.RegularExpressions;

namespace Platform.Extensions
{
    /// <summary>
    /// Conversion without exceptions and 
    /// with default values
    /// </summary>
    public static class ConvertExtensions
    {
        #region Type Conversion String Parsers

        #region Radix decode and Radix encode

        // valid radixes
        private static string _AplhaDecimal = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
        // valid radixes
        private static string _SafeAplhaDecimal = "0123456789ABCDEFGHJKLMNPQRTUVWXYZ";

        /// <summary>
        /// Convert string to value
        /// </summary>
        /// <param name="source"></param>
        /// <param name="radixes"></param>
        /// <returns></returns>
        public static decimal? RadixDecode(this string source, string radix)
        {
            // set value
            decimal value = 0;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radix.IndexOf(source[index]);
                if (key == -1) return value;

                // check range
                var next = value * radix.Length + key;
                if (next < value) return null;
                
                // check value
                value = next;
            }

            // return value
            return value;
        }

        /// <summary>
        /// Convert value to alpha decimal string
        /// </summary>
        /// <param name="value"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string RadixEncode(this decimal value, int length, string radix)
        {
            // create result
            StringBuilder result = new StringBuilder();

            var radixes = radix.ToCharArray();

            // convert to string
            for (int i = 0; (length!=-1 && i<length) || (length==-1 && value!=0) || result.Length==0; i++)
            {                
                var fraction = (int)Decimal.Truncate( value % radix.Length );
                value = Decimal.Truncate( value / radix.Length );
                result.Insert(0, radixes, fraction, 1);
            }

            // return result as string
            return result.ToString();
        }

        #endregion

        #region To SByte

        /// <summary>
        /// Safe Parse SByte
        /// </summary>
        public static SByte? ToSByte(this string source) { return ToSByte(source, 10, null); }

        /// <summary>
        /// Safe Parse SByte
        /// </summary>
        public static SByte? ToSByte(this string source, int radix) { return ToSByte(source, radix, null); }

        /// <summary>
        /// Safe Parse SByte
        /// </summary>
        public static SByte? ToSByte(this string source, int radix, SByte? value) { return ToSByte(source, _AplhaDecimal.Substring(0, radix), value); }

        /// <summary>
        /// Safe Parse SByte
        /// </summary>
        public static SByte? ToSByte(this string source, string radixes, SByte? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // prepare result
            SByte result = 0;
            var limit = source[0] == '-' ? SByte.MinValue : SByte.MaxValue;

            // convert to valid string
            if (source[0] == '-' || source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;
                key *= Math.Sign(limit);

                // check sign
                if (Math.Abs(result) > Math.Abs((limit - key) / radixes.Length)) return value;

                // update result
                result = (SByte)(result * radixes.Length);
                result += (SByte)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To Int16

        /// <summary>
        /// Safe Parse Int16
        /// </summary>
        public static Int16? ToInt16(this string source) { return ToInt16(source, 10, null); }

        /// <summary>
        /// Safe Parse Int16
        /// </summary>
        public static Int16? ToInt16(this string source, int radix) { return ToInt16(source, radix, null); }

        /// <summary>
        /// Safe Parse Int16
        /// </summary>
        public static Int16? ToInt16(this string source, int radix, Int16? value) { return ToInt16(source, _AplhaDecimal.Substring(0, radix), value); }

        /// <summary>
        /// Safe Parse Int16
        /// </summary>
        public static Int16? ToInt16(this string source, string radixes, Int16? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // prepare result
            Int16 result = 0;
            var limit = source[0] == '-' ? Int16.MinValue : Int16.MaxValue;

            // convert to valid string
            if (source[0] == '-' || source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;
                key *= Math.Sign(limit);

                // check sign
                if (Math.Abs(result) > Math.Abs((limit - key) / radixes.Length)) return value;

                // update result
                result = (Int16)(result * radixes.Length);
                result += (Int16)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To Int32

        /// <summary>
        /// Safe Parse Int32
        /// </summary>
        public static Int32? ToInt32(this string source) { return ToInt32(source, 10, null); }

        /// <summary>
        /// Safe Parse Int32
        /// </summary>
        public static Int32? ToInt32(this string source, int radix) { return ToInt32(source, radix, null); }

        /// <summary>
        /// Safe Parse Int32
        /// </summary>
        public static Int32? ToInt32(this string source, int radix, Int32? value) { return ToInt32(source, _AplhaDecimal.Substring(0, radix), value); }

        /// <summary>
        /// Safe Parse Int32
        /// </summary>
        public static Int32? ToInt32(this string source, string radixes, Int32? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // ignore seperators
            source = source.Replace(".", "");
            source = source.Replace(",", "");

            // prepare result
            Int32 limit = source[0] == '-' ? Int32.MinValue : Int32.MaxValue;
            Int32 result = 0;

            // convert to valid string
            if (source[0] == '-' || source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;
                key *= Math.Sign(limit);

                // check sign
                if (Math.Abs(result) > Math.Abs((limit - key) / radixes.Length)) return value;

                // update result
                result = (Int32)(result * radixes.Length);
                result += (Int32)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To Int64

        /// <summary>
        /// Safe Parse Int64
        /// </summary>
        public static Int64? ToInt64(this string source) { return ToInt64(source, 10, null); }

        /// <summary>
        /// Safe Parse Int64
        /// </summary>
        public static Int64? ToInt64(this string source, int radix) { return ToInt64(source, radix, null); }

        /// <summary>
        /// Safe Parse Int64
        /// </summary>
        public static Int64? ToInt64(this string source, int radix, Int64? value) { return ToInt64(source, _AplhaDecimal.Substring(0, radix), value); }

        /// <summary>
        /// Safe Parse Int64
        /// </summary>
        public static Int64? ToInt64(this string source, string radixes, Int64? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // prepare result
            Int64 limit = source[0] == '-' ? Int64.MinValue : Int64.MaxValue;
            Int64 result = 0;

            // convert to valid string
            if (source[0] == '-' || source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;
                key *= Math.Sign(limit);

                // check sign
                if (Math.Abs(result) > Math.Abs((limit - key) / radixes.Length)) return value;

                // update result
                result = (Int64)(result * radixes.Length);
                result += (Int64)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To Byte

        /// <summary>
        /// Safe Parse Int32
        /// </summary>
        public static Byte? ToByte(this string source) { return ToByte(source, 10, null); }

        /// <summary>
        /// Safe Parse Int32
        /// </summary>
        public static Byte? ToByte(this string source, int radix) { return ToByte(source, radix, null); }

        /// <summary>
        /// Safe Parse Int32
        /// </summary>
        public static Byte? ToByte(this string source, int radix, Byte? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // prepare parsing
            string radixes = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ".Substring(0, radix);

            // prepare result
            Byte limit = Byte.MaxValue;
            Byte result = 0;

            // convert to valid string
            if (source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;

                // check sign
                if (result > (limit - key) / radix) return value;

                // update result
                result = (Byte)(result * radix);
                result += (Byte)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To UInt16

        /// <summary>
        /// Safe Parse UInt16
        /// </summary>
        public static UInt16? ToUInt16(this string source) { return ToUInt16(source, 10, null); }

        /// <summary>
        /// Safe Parse UInt16
        /// </summary>
        public static UInt16? ToUInt16(this string source, int radix) { return ToUInt16(source, radix, null); }

        /// <summary>
        /// Safe Parse UInt16
        /// </summary>
        public static UInt16? ToUInt16(this string source, int radix, UInt16? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // prepare parsing
            string radixes = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ".Substring(0, radix);

            // prepare result
            UInt16 limit = UInt16.MaxValue;
            UInt16 result = 0;

            // convert to valid string
            if (source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;

                // check sign
                if (result > (limit - key) / radix) return value;

                // update result
                result = (UInt16)(result * radix);
                result += (UInt16)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To UInt32

        /// <summary>
        /// Safe Parse UInt32
        /// </summary>
        public static UInt32? ToUInt32(this string source) { return ToUInt32(source, 10, null); }

        /// <summary>
        /// Safe Parse UInt32
        /// </summary>
        public static UInt32? ToUInt32(this string source, int radix) { return ToUInt32(source, radix, null); }

        /// <summary>
        /// Safe Parse UInt32
        /// </summary>
        public static UInt32? ToUInt32(this string source, int radix, UInt32? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // prepare parsing
            string radixes = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ".Substring(0, radix);

            // prepare result
            UInt32 limit = UInt32.MaxValue;
            UInt32 result = 0;

            // convert to valid string
            if (source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;

                // check sign
                if (result > (limit - key) / radix) return value;

                // update result
                result = (UInt16)(result * radix);
                result += (UInt16)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To UInt64

        /// <summary>
        /// Safe Parse UInt64
        /// </summary>
        public static UInt64? ToUInt64(this string source) { return ToUInt64(source, 10, null); }

        /// <summary>
        /// Safe Parse UInt64
        /// </summary>
        public static UInt64? ToUInt64(this string source, int radix) { return ToUInt64(source, radix, null); }

        /// <summary>
        /// Safe Parse UInt64
        /// </summary>
        public static UInt64? ToUInt64(this string source, int radix, UInt64? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // prepare parsing
            string radixes = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ".Substring(0, radix);

            // prepare result
            UInt64 limit = UInt32.MaxValue;
            UInt64 result = 0;

            // convert to valid string
            if (source[0] == '+') source = source.Substring(1).ToUpper();
            else source = source.ToUpper();

            // must have some characters
            if (source.Length == 0) return value;

            // parse all characters
            for (int index = 0; index < source.Length; index++)
            {
                // get index
                int key = radixes.IndexOf(source[index]);
                if (key == -1) return value;

                // check sign
                if (result > (limit - (UInt64)key) / (UInt64)radix) return value;

                // update result
                result = (UInt64)(result * (UInt64)radix);
                result += (UInt64)key;
            }

            // return result
            return result;
        }

        #endregion

        #region To Float

        /// <summary>
        /// Safe Parse Single
        /// </summary>
        public static Single? ToFloat(this string source) { return ToFloat(source, ",.", null); }

        /// <summary>
        /// Safe Parse Single
        /// </summary>
        public static Single? ToFloat(this string source, string sep) { return ToFloat(source, sep, null); }

        /// <summary>
        /// Safe Parse Single
        /// </summary>
        public static Single? ToFloat(this string source, string sep, Single? value)
        {
            return (Single?)ToDouble(source, sep, value);
        }

        #endregion

        #region To Double

        /// <summary>
        /// Safe Parse Double
        /// </summary>
        public static Double? ToDouble(this string source) { return ToDouble(source, ".,", null); }

        /// <summary>
        /// Safe Parse Double
        /// </summary>
        public static Double? ToDouble(this string source, string sep) { return ToDouble(source, sep, null); }

        /// <summary>
        /// Safe Parse Double
        /// </summary>
        public static Double? ToDouble(this string source, string sep, Double? value)
        {
            // must have some characters
            if (source == null || source.Length == 0) return value;

            // check sign
            bool negative = false;
            int sign = source.IndexOfAny(new char[] { '-', '+' });
            if (sign != -1)
            {
                negative = source[sign] == '-';
                source.Remove(sign, 1);
            }

            // must have some characters
            if (source.Length == 0) return value;

            // check for multiple or single use of seperator
            int seppos = -1;
            for (int i = 0; i < sep.Length; i++)
                if (source.ToCharArray().Count(c => c == sep[i]) == 1)
                    seppos = source.IndexOf(sep[i]);

            // default to int
            string main = source;
            string subs = "";

            // remove seperators
            if (seppos != -1)
            {
                main = source.Substring(0, seppos).Filter(sep.ToCharArray());
                subs = source.Substring(seppos+1).Filter(sep.ToCharArray());
            }

            // remove ignored seperators
            if (subs.Length == 0 && main.Length == 0) return value;

            // get integer part
            UInt64? part1 = ToUInt64(main, 10, null);
            if (part1 == null && main.Length != 0) return value;

            // get real part
            UInt64? part2 = ToUInt64(subs, 10, null);
            if (part2 == null && subs.Length != 0) return value;

            // get result
            Single result = 0;
            if (main.Length > 0 && part1 != null) result += part1.Value;
            if (subs.Length > 0 && part2 != null) result += (Single)(part2.Value / Math.Pow(10.0, subs.Length));

            // apply sign
            if (negative) result = -result;

            // return result
            return result;
        }

        #endregion

        #region ToEnum

        /// <summary>
        /// Enum conversion
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="defaultvalue"></param>
        /// <returns></returns>
        public static T ToEnum<T>(this string source, T defaultvalue = default(T) )
            where T : struct
        {
            T value;
            if (!System.Enum.TryParse<T>(source, false, out value)) value = defaultvalue;
            return value;
        }

        #endregion

        #region ToBoolean

        /// <summary>
        /// Enum conversion
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="defaultvalue"></param>
        /// <returns></returns>
        public static Boolean? ToBoolean(this string source, Boolean? defaultvalue)
        {
            if (source.IsEqualTo("true")) return true;
            if (source.IsEqualTo("false")) return false;
            if (source.IsEqualTo("yes")) return true;
            if (source.IsEqualTo("no")) return false;
            if (source.IsEqualTo("ja")) return true;
            if (source.IsEqualTo("nee")) return false;
            if (source.IsEqualTo("1")) return true;
            if (source.IsEqualTo("0")) return false;
            return defaultvalue;
        }

        #endregion

        #region ToDateTime (TODO: ignore exceptions)

        /// <summary>
        /// Safe Parse DateTime
        /// </summary>
        public static DateTime? ToDateTime(this string source) { return ToDateTime(source, null); }

        /// <summary>
        /// Safe Parse DateTime
        /// </summary>
        public static DateTime? ToDateTime(this string source, string format) { return ToDateTime(source, format, null); }

        /// <summary>
        /// Safe Parse DateTime
        /// </summary>
        public static DateTime? ToDateTime(this string source, string format, DateTime? value)
        {
            try
            {
                var flags = (format.ToUpper()=="O" || format.ToUpper()=="R") ? 
                    DateTimeStyles.RoundtripKind : DateTimeStyles.None;
                if (format == null) value = DateTime.Parse(source, null, flags);
                else value = DateTime.ParseExact(source, format, null, flags);

            }
            catch
            {
            }
            return value;
        }

        #endregion

        /// <summary>
        /// Parse Value
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        public static TimeSpan? ToTimeSpan( this string source, TimeSpan? defaultValue = null )
        {
            // check source valid
            if (source == null)
                return defaultValue;

            // split parts
            var parts = source.Split(':');

            // check if hh:mm:ss
            if ( parts.Length == 3)
            {
                var hour = parts[0].ToInt32();
                if (hour == null) return defaultValue;
                var min = parts[0].ToInt32();
                if (min == null) return defaultValue;
                var sec = parts[1].ToInt32();
                if (sec == null) return defaultValue;
                return new TimeSpan(hour.Value, min.Value, sec.Value);
            }

            // check if mm:ss
            if (parts.Length == 2)
            {
                var min = parts[0].ToInt32();
                if (min == null) return defaultValue;
                var sec = parts[1].ToInt32();
                if (sec == null) return defaultValue;
                return new TimeSpan(0, min.Value, sec.Value);
            }

            if (parts.Length == 1 && source.EndsWith("d"))
            {
                var val = source.Substring(0, source.Length - 1).ToDouble();
                if (val == null) return defaultValue;
                return TimeSpan.FromDays(val.Value);
            }

            if (parts.Length == 1 && (source.EndsWith("h") || source.EndsWith("u")) )
            {
                var val = source.Substring(0, source.Length - 1).ToDouble();
                if (val == null) return defaultValue;
                return TimeSpan.FromHours(val.Value);
            }

            if (parts.Length == 1 && source.EndsWith("m"))
            {
                var val = source.Substring(0, source.Length - 1).ToDouble();
                if (val == null) return defaultValue;
                return TimeSpan.FromMinutes(val.Value);
            }

            if (parts.Length == 1 && source.EndsWith("s"))
            {
                var val = source.Substring(0, source.Length - 1).ToDouble();
                if (val == null) return defaultValue;
                return TimeSpan.FromSeconds(val.Value);
            }

            if (parts.Length == 1 )
            {
                var val = source.ToInt32();
                if (val == null) return defaultValue;
                return TimeSpan.FromMilliseconds(val.Value);
            }

            // check if hh:mm:ss
            return defaultValue;
                
        }

        #endregion

        #region ToString radix

        /// <summary>
        /// Convert value to alpha decimal string
        /// </summary>
        /// <param name="value"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string ToAlphaDecimal(this long value, int length)
        {
            return RadixEncode(value, length, _AplhaDecimal);
        }

        /// <summary>
        /// Convert value to alpha decimal string
        /// </summary>
        /// <param name="value"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string ToAlphaDecimal(this int value, int length)
        {
            return RadixEncode(value, length, _AplhaDecimal);
        }

        /// <summary>
        /// Convert value to alpha decimal string
        /// </summary>
        /// <param name="value"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static decimal? FromAlphaDecimal(this string value)
        {
            return RadixDecode( value, _AplhaDecimal);
        }

        /// <summary>
        /// Convert value to alpha decimal string
        /// </summary>
        /// <param name="value"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string ToSafeAlphaDecimal(this long value, int length)
        {
            return RadixEncode(value, length, _SafeAplhaDecimal);
        }

        /// <summary>
        /// Convert value to alpha decimal string
        /// </summary>
        /// <param name="value"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static decimal? FromSafeAlphaDecimal(this string value)
        {
            return RadixDecode(value, _SafeAplhaDecimal);
        }

        #endregion

        #region Bit Converters For Network Byte Order

        /// <summary>
        /// Read ulong value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static long ReadNetworkInt64(this byte[] data, int offset)
        {
            unchecked
            {
                ulong value = 0;
                value |= (ulong)data[offset + 0] << 56;
                value |= (ulong)data[offset + 1] << 48;
                value |= (ulong)data[offset + 2] << 40;
                value |= (ulong)data[offset + 3] << 32;
                value |= (ulong)data[offset + 4] << 24;
                value |= (ulong)data[offset + 5] << 16;
                value |= (ulong)data[offset + 6] << 8;
                value |= (ulong)data[offset + 7] << 0;
                return (long)value;
            }
        }

        /// <summary>
        /// Read uint value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static int ReadNetworkInt32(this byte[] data, int offset)
        {
            unchecked
            {
                uint value = 0;
                value |= (uint)data[offset + 0] << 24;
                value |= (uint)data[offset + 1] << 16;
                value |= (uint)data[offset + 2] << 8;
                value |= (uint)data[offset + 3] << 0;
                return (int)value;
            }
        }

        /// <summary>
        /// Read uint value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static int ReadNetworkInt24(this byte[] data, int offset)
        {
            unchecked
            {
                uint value = 0;
                value |= (uint)data[offset + 0] << 16;
                value |= (uint)data[offset + 1] << 8;
                value |= (uint)data[offset + 2] << 0;
                return (int)value;
            }
        }

        /// <summary>
        /// Read uint value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static short ReadNetworkInt16(this byte[] data, int offset)
        {
            unchecked
            {
                uint value = 0;
                value |= (uint)data[offset + 0] << 8;
                value |= (uint)data[offset + 1] << 0;
                return (short)value;
            }
        }

        /// <summary>
        /// Read ulong value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static ulong ReadNetworkUInt64(this byte[] data, int offset)
        {
            unchecked
            {
                ulong value = 0;
                value |= (ulong)data[offset + 0] << 56;
                value |= (ulong)data[offset + 1] << 48;
                value |= (ulong)data[offset + 2] << 40;
                value |= (ulong)data[offset + 3] << 32;
                value |= (ulong)data[offset + 4] << 24;
                value |= (ulong)data[offset + 5] << 16;
                value |= (ulong)data[offset + 6] << 8;
                value |= (ulong)data[offset + 7] << 0;
                return value;
            }
        }

        /// <summary>
        /// Read uint value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static uint ReadNetworkUInt32(this byte[] data, int offset)
        {
            unchecked
            {
                uint value = 0;
                value |= (uint)data[offset + 0] << 24;
                value |= (uint)data[offset + 1] << 16;
                value |= (uint)data[offset + 2] << 8;
                value |= (uint)data[offset + 3] << 0;
                return (uint)value;
            }
        }

        /// <summary>
        /// Read uint value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static int ReadNetworkUInt24(this byte[] data, int offset)
        {
            unchecked
            {
                uint value = 0;
                value |= (uint)data[offset + 0] << 16;
                value |= (uint)data[offset + 1] << 8;
                value |= (uint)data[offset + 2] << 0;
                return (int)value;
            }
        }

        /// <summary>
        /// Read uint value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static ushort ReadNetworkUInt16(this byte[] data, int offset)
        {
            unchecked
            {
                uint value = 0;
                value |= (uint)data[offset + 0] << 8;
                value |= (uint)data[offset + 1] << 0;
                return (ushort)value;
            }
        }

        /// <summary>
        /// Read network value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <param name="count"></param>
        /// <returns></returns>
        public static ulong ReadNetworkValue(this byte[] data, ref int offset, int count)
        {
            ulong value = 0;

            for (int i = 0; i < count; i++)
            {
                value += data[offset++];
                value *= 256;
            }

            return value;
        }

        /// <summary>
        /// write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkInt64(this byte[] data, int offset, long value)
        {
            data[offset + 0] = (byte)unchecked((value >> 56) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 48) & 0xFF);
            data[offset + 2] = (byte)unchecked((value >> 40) & 0xFF);
            data[offset + 3] = (byte)unchecked((value >> 32) & 0xFF);
            data[offset + 4] = (byte)unchecked((value >> 24) & 0xFF);
            data[offset + 5] = (byte)unchecked((value >> 16) & 0xFF);
            data[offset + 6] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 7] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkInt32(this byte[] data, int offset, int value)
        {
            data[offset + 0] = (byte)unchecked((value >> 24) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 16) & 0xFF);
            data[offset + 2] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 3] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkInt24(this byte[] data, int offset, int value)
        {
            data[offset + 0] = (byte)unchecked((value >> 16) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 2] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// Write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkInt16(this byte[] data, int offset, int value)
        {
            data[offset + 0] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkUInt64(this byte[] data, int offset, ulong value)
        {
            data[offset + 0] = (byte)unchecked((value >> 56) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 48) & 0xFF);
            data[offset + 2] = (byte)unchecked((value >> 40) & 0xFF);
            data[offset + 3] = (byte)unchecked((value >> 32) & 0xFF);
            data[offset + 4] = (byte)unchecked((value >> 24) & 0xFF);
            data[offset + 5] = (byte)unchecked((value >> 16) & 0xFF);
            data[offset + 6] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 7] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkUInt32(this byte[] data, int offset, long value)
        {
            data[offset + 0] = (byte)unchecked((value >> 24) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 16) & 0xFF);
            data[offset + 2] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 3] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkUInt24(this byte[] data, int offset, int value)
        {
            data[offset + 0] = (byte)unchecked((value >> 16) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 2] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// Write value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static void WriteNetworkUInt16(this byte[] data, int offset, int value)
        {
            data[offset + 0] = (byte)unchecked((value >> 8) & 0xFF);
            data[offset + 1] = (byte)unchecked((value >> 0) & 0xFF);
        }

        /// <summary>
        /// Write network value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <param name="count"></param>
        /// <returns></returns>
        public static byte[] WriteNetworkValue(ulong value, int minbytes)
        {
            // create target buffer
            int offset = 0;
            byte[] buffer = new byte[8];

            // write bytes
            for (; value > 0 || minbytes > 0; minbytes--)
            {
                buffer[offset++] = (byte)(value % 256);
                value /= 256;
            }

            // resize
            Array.Resize(ref buffer, offset);

            // reverse
            Array.Reverse(buffer);

            // return result
            return buffer;
        }

        #endregion

        #region Read & Write Null terminated ASCII string from/to byte array

#if !SILVERLIGHT && !NETPL

        /// <summary>
        /// Gets an ascii string
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string GetNullTerminatedAsciiString(this byte[] data, int offset, int length)
        {
            // get size
            int size;
            for (size = 0; size < length; size++)
                if (data[offset + size] == 0) break;

            // return string
            return Encoding.ASCII.GetString(data, offset, size);
        }

        // Sets an ascii string
        public static void SetNullTerminatedAsciiString(this byte[] data, int offset, int length, string value)
        {
            var bytes = Encoding.ASCII.GetBytes(value);
            Array.Copy(bytes, 0, data, offset, Math.Max(length, bytes.Length));
            data[offset + length - 1] = 0;
        }

#endif

        #endregion

        #region Read & Write Length Prefix String from/to byte array

        /// <summary>
        /// Read string value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public static string ReadLengthString(byte[] data, ref int offset, Encoding encoding)
        {
            int length = (int)data[offset++];
            string value = encoding.GetString(data, offset, length);
            offset += length;
            return value;
        }

        /// <summary>
        /// Write string value
        /// </summary>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        public static void WriteLengthString(byte[] data, ref int offset, string value, Encoding encoding)
        {
            if (value == null) return;
            byte[] text = encoding.GetBytes(value);
            data[offset++] = (byte)text.Length;
            Array.Copy(text, 0, data, offset, text.Length);
            offset += text.Length;
        }

        #endregion

        #region Type Conversion byte to hex

        /// <summary>
        /// Convert to hex string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="offset"></param>
        /// <param name="count"></param>
        /// <returns></returns>
        public static string ToStringHex(this byte[] source, int offset, int count)
        {
            StringBuilder result = new StringBuilder();
            for (int i = 0; i < count; i++, offset++) result.AppendFormat("{0:X02}", source[offset]);
            return result.ToString();
        }

        /// <summary>
        /// Convert to hex string
        /// </summary>
        /// <param name="source"></param>
        /// <returns></returns>
        public static string ToStringHex(this byte[] source)
        {
            return ToStringHex(source, 0, source.Length);
        }

        /// <summary>
        /// Covert hex string to byte[]
        /// </summary>
        /// <param name="source"></param>
        /// <returns></returns>
        public static byte[] ToByteArrayHex( this string source )
        {
            byte[] target = new byte[source.Length / 2];
            for( int i=0; i<source.Length; i+=2 )
                target[i / 2] = source.Substring(i, 2).ToByte(16)??0;
            return target;
        }

        #endregion

        #region DateTimeToNTP32

        /// <summary>
        /// Converts specified DateTime value to short NTP time. Note: NTP time is in UTC.
        /// </summary>
        /// <param name="value">DateTime value to convert. This value must be in local time.</param>
        /// <returns>Returns NTP value.</returns>
        public static uint DateTimeToNTP32(DateTime value)
        {
            unchecked
            {
                TimeSpan span = value.ToUniversalTime() - new DateTime(1900, 1, 1, 0, 0, 0);
                ushort seconds = (ushort)span.TotalSeconds;
                ushort fraction = (ushort)((span.TotalSeconds - seconds) * ushort.MaxValue);
                return (uint)seconds << 16 | (uint)fraction;
            }
        }

        #endregion

        #region DateTimeToNTP64

        /// <summary>
        /// Converts specified DateTime value to long NTP time. Note: NTP time is in UTC.
        /// </summary>
        /// <param name="value">DateTime value to convert. This value must be in local time.</param>
        /// <returns>Returns NTP value.</returns>
        public static ulong DateTimeToNTP64(DateTime value)
        {
            unchecked
            {
                TimeSpan span = value.ToUniversalTime() - new DateTime(1900, 1, 1, 0, 0, 0);
                uint seconds = (uint)span.TotalSeconds;
                uint fraction = (uint)((span.TotalSeconds - seconds) * uint.MaxValue);
                return (ulong)seconds << 32 | (ulong)fraction;
            }
        }

        #endregion

        #region Guid To String

        /// <summary>
        /// Convert guid to compact string (upper case no dashes)
        /// </summary>
        /// <param name="guid"></param>
        /// <returns></returns>
        public static string ToCompactString(this Guid guid)
        {
            return guid.ToString().Replace("-", "").ToUpper();
        }

        #endregion

        #region Physical Address

        public static UInt64? PhysicalAddressToLong(string address, UInt64? @default)
        {
            if (address == null)
                return @default;

            address = address
                .Replace(":", "")
                .Replace("-", "")
                .Replace(".", "");

            return address.ToUInt64(16, @default);
        }

        public static string LongToPhysicalAddress(UInt64 address)
        {
            var result = string.Format( "{0:X012}", address & 0xFFFFFFFFFFFF );
            return result.Substring(0, 2) + ":" +
                   result.Substring(2, 2) + ":" +
                   result.Substring(4, 2) + ":" +
                   result.Substring(6, 2) + ":" +
                   result.Substring(8, 2) + ":" +
                   result.Substring(10, 2);
        }

        #endregion
    }
}
