﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;

namespace Platform.Extensions
{
    public static class EnumExtensions
    {
        /// <summary>
        /// Parse enum from string
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="default"></param>
        /// <returns></returns>
        public static T Parse<T>(this string source, T @default) where T : struct
        {
            // get enum type
            var type = typeof(T);
            if (!type.IsEnum) throw new ArgumentException("Expected an Enum");

            // get fields
            var field = type.GetFields().FirstOrDefault(f => f.Name.IsEqualTo(source));
            if (field == null) return @default;

            // return enum
            var value = field.GetValue(null);
            return (T)value;
        }

        /// <summary>
        /// Test if enumeration contains a flag
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="type"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static bool Has<T>(this System.Enum type, T value)
        {
            try
            {
                return (((int)(object)type & (int)(object)value) == (int)(object)value);
            }
            catch
            {
                return false;
            }
        }

        /// <summary>
        /// Check if enumeration equals
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="type"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static bool Is<T>(this System.Enum type, T value)
        {
            try
            {
                return (int)(object)type == (int)(object)value;
            }
            catch
            {
                return false;
            }
        }

        /// <summary>
        /// Add flag to enumeration
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="type"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static T Add<T>(this System.Enum type, T value)
        {
            try
            {
                return (T)(object)(((int)(object)type | (int)(object)value));
            }
            catch (Exception ex)
            {
                string prompt = string.Format( "Could not append value from enumerated type '{0}'.", typeof(T).Name );
                throw new ArgumentException( prompt, ex);
            }
        }

        /// <summary>
        /// Remove flag from enumeration
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="type"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static T Remove<T>(this System.Enum type, T value)
        {
            try
            {
                return (T)(object)(((int)(object)type & ~(int)(object)value));
            }
            catch (Exception ex)
            {
                string prompt = string.Format("Could not remove value from enumerated type '{0}'.", typeof(T).Name);
                throw new ArgumentException(prompt, ex);
            }
        }

        public static IEnumerable<T> GetValues<T>()
        {
            return typeof(T)
                .GetFields()
                .Where(x => x.IsLiteral)
                .Select(field => (T)field.GetValue(null));
        }
    } 
}
