﻿using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Collections.Generic;

using System.Net.Sockets;

namespace Platform.Extensions
{
    public static partial class NetworkInterfaceExtensions
    {
        /// <summary>
        /// Gets all IP addresses
        /// </summary>
        /// <returns></returns>
        public static IPAddress[] GetAllIpAddresses()
        {
            List<IPAddress> addresses = new List<IPAddress>();
            foreach (var nic in NetworkInterface.GetAllNetworkInterfaces())
            {
                addresses.AddRange(nic.GetIPProperties().UnicastAddresses.ToArray().Select(t => t.Address));
                addresses.AddRange(nic.GetIPProperties().AnycastAddresses.ToArray().Select(t => t.Address));
                addresses.AddRange(nic.GetIPProperties().MulticastAddresses.ToArray().Select(t => t.Address));
            }
            return addresses.ToArray();
        }

        /// <summary>
        /// Returns all bound IP addresses
        /// </summary>
        /// <param name="nic"></param>
        /// <returns></returns>
        public static IPAddress[] GetAllIpAddresses(this NetworkInterface nic)
        {
            List<IPAddress> addresses = new List<IPAddress>();
            addresses.AddRange(nic.GetIPProperties().UnicastAddresses.ToArray().Select(t => t.Address));
            addresses.AddRange(nic.GetIPProperties().AnycastAddresses.ToArray().Select(t => t.Address));
            addresses.AddRange(nic.GetIPProperties().MulticastAddresses.ToArray().Select(t => t.Address));
            return addresses.ToArray();
        }

        /// <summary>
        /// Get physical address from IP address
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public static IEnumerable<UnicastIPAddressInformation> GetUnicastAddresses()
        {
            foreach (var nic in NetworkInterface.GetAllNetworkInterfaces())
                foreach (var ad in nic.GetIPProperties().UnicastAddresses)
                    yield return ad;
        }

        /// <summary>
        /// Get physical address from IP address
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public static PhysicalAddress GetPhysicalAddressFromIpAddress(IPAddress address)
        {
            foreach (var nic in NetworkInterface.GetAllNetworkInterfaces())
                foreach (var ad in nic.GetIPProperties().UnicastAddresses)
                    if (ad.Address.AddressFamily == AddressFamily.InterNetwork)
                        if (ad.Address.Equals(address))
                            return nic.GetPhysicalAddress();
            return null;
        }

        /// <summary>
        /// Get physical address from IP address
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public static NetworkInterface GetInterfaceFromIpAddress(IPAddress address)
        {
            foreach (var nic in NetworkInterface.GetAllNetworkInterfaces())
                foreach (var ad in nic.GetIPProperties().UnicastAddresses)
                    if (ad.Address.AddressFamily == address.AddressFamily)
                        if (ad.Address.Equals(address))
                            return nic;
            return null;
        }

        /// <summary>
        /// Update local endpoint, if possible, to reflect the 
        /// address for the AdapterName, AdapterPort and AdapterType
        /// </summary>
        public static IPAddress GetIpAddressFromInterface(string name, AddressFamily? family)
        {
            // adapter must exist
            var nic = NetworkInterface.GetAllNetworkInterfaces().FirstOrDefault(n => n.Name == name);
            if (nic == null) return IPAddress.None;

            // get all addresses for nic
            var addresses = nic.GetIPProperties().UnicastAddresses.Select(a => a.Address).Where(a => family == null || a.AddressFamily == family.Value);

            // find first available valid IP address
            var address = addresses.FirstOrDefault(a => a.IsValidForBinding());
            if (address == null) return IPAddress.None;

            // return found address
            return address;
        }

        /// <summary>
        /// Check if address is within subnet
        /// </summary>
        /// <param name="addressinfo"></param>
        /// <param name="address"></param>
        /// <returns></returns>
        public static bool IsAddressWithinSubnet(this UnicastIPAddressInformation network, IPAddress address)
        {
            // check address family
            if (address.AddressFamily != network.Address.AddressFamily) return false;

            byte[] filter = null;
            var target = address.GetAddressBytes();
            var source = network.Address.GetAddressBytes();

            // check IPv4
            if (address.AddressFamily == AddressFamily.InterNetwork)
                filter = network.IPv4Mask.GetAddressBytes();

            // check IPv6
            if (address.AddressFamily == AddressFamily.InterNetworkV6)
                filter = new byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0, 0, 0, 0, 0, 0, 0, 0 };

            // invalid address type
            if( filter == null ) 
                return false;

            // check if subnet the same
            for( int i=0; i<target.Length; i++ )
                if( (source[i] & filter[i]) != (target[i] & filter[i]) )
                    return false;

            // return OK
            return true;
        }
    }
}
