﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;

namespace Platform.Extensions
{
    public static class ObjectExtensions
    {
        /// <summary>
        /// Gets a dynamic value
        /// </summary>
        /// <param name="item"></param>
        /// <param name="name"></param>
        /// <returns></returns>
        public static object GetPropertyValue(this object item, string name, params object[] indexes)
        {
            while (true)
            {
                // parse path
                string property = name;
                int index = name.IndexOf('.');
                if (index == -1) name = "";
                else
                {
                    property = name.Substring(0,index);
                    name = name.Substring(index + 1);
                }

                // check if property was given
                if (string.IsNullOrEmpty(property))
                    return item;

                // null reference
                if (item == null) 
                    throw new NullReferenceException();

                // get class type                
                var type = (item is Type) ? (Type)item : item.GetType();
                var flags = (item is Type) ? BindingFlags.Static : BindingFlags.Instance;

                // get property info
                var info = type
                    .GetProperties(flags|BindingFlags.Public|BindingFlags.NonPublic)
                    .Concat(type.GetInterfaces().SelectMany(i => i.GetProperties()) )
                    .FirstOrDefault(p => p.Name == property && p.GetIndexParameters().Length == 0);
                if (info == null) throw new MissingMemberException(string.Format( "The property {1} was not found on type {0}", type.Name, property ));

                // get property value
                item = info.GetValue(item, indexes);
            }
        }

        /// <summary>
        /// Gets a dynamic value
        /// </summary>
        /// <param name="item"></param>
        /// <param name="name"></param>
        /// <returns></returns>
        public static object GetFieldValue(this object item, string name, params object[] indexes)
        {
            while (true)
            {
                // parse path
                string field = name;
                int index = name.IndexOf('.');
                if (index == -1) name = "";
                else
                {
                    field = name.Substring(0, index);
                    name = name.Substring(index + 1);
                }

                // check if property was given
                if (string.IsNullOrEmpty(field))
                    return item;

                // null reference
                if (item == null)
                    throw new NullReferenceException();

                // get class type                
                var type = (item is Type) ? (Type)item : item.GetType();
                var flags = (item is Type) ? BindingFlags.Static : BindingFlags.Instance;

                // get property info
                var info = type.GetField(field, flags | BindingFlags.Public | BindingFlags.NonPublic);
                if (info == null) throw new MissingMemberException(string.Format("The property {1} was not found on type {0}", type.Name, field));

                // get property value
                item = info.GetValue(item);
            }
        }

        /// <summary>
        /// Set dynamic property
        /// </summary>
        /// <param name="item"></param>
        /// <param name="name"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        public static void SetPropertyValue(this object item, string name, object value, params object[] indexes)
        {
            // check for path name
            int index = name.LastIndexOf('.');
            if (index != -1)
            {
                item = GetPropertyValue(item, name.Substring(0, index));
                name = name.Substring(index + 1);
            }

            // null reference
            if (item == null)
                throw new NullReferenceException();

            // get class type                
            var type = (item is Type) ? (Type)item : item.GetType();
            var flags = (item is Type) ? BindingFlags.Static : BindingFlags.Instance;

            // get method info
            var info = type.GetProperty(name, flags | BindingFlags.Public | BindingFlags.NonPublic);
            if (info == null) throw new MissingMemberException(string.Format("The property {1} was not found on type {0}", type.Name, name));

            // invoke metod
            info.SetValue(item, value, indexes);
        }

        /// <summary>
        /// Invoke dynamic method
        /// </summary>
        /// <param name="item"></param>
        /// <param name="name"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        public static object InvokeMethod(this object item, string name, params object[] args)
        {
            // check for path name
            int index = name.LastIndexOf( '.' );
            if( index != -1 )
            {
                item = GetPropertyValue( item, name.Substring( 0, index ) );
                name = name.Substring( index + 1 );
            }

            // null reference
            if (item == null) 
                throw new NullReferenceException();

            // get class type                
            var type = (item is Type) ? (Type)item : item.GetType();
            var flags = (item is Type) ? BindingFlags.Static : BindingFlags.Instance;

            while( true )
            {
                // get method info
                var info = type.GetMethod(name, flags | BindingFlags.Public | BindingFlags.NonPublic);

                if (info != null)
                    return info.Invoke(item, args);

                if (type.BaseType == null )
                    throw new MissingMethodException(string.Format("The method {1} was not found on type {0}", type.Name, name));

                type = type.BaseType;
            }
        }
    }
}
