﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;
using System.Xml.Linq;

namespace Platform.Extensions
{
    public static class XmlConvert
    {
        /// <summary>
        /// Covert from string to type using Xml
        /// </summary>
        /// <param name="source"></param>
        /// <param name="type"></param>
        /// <returns></returns>
        public static object Convert(string source, Type type)
        {
            var name = type.Name;

            if (Nullable.GetUnderlyingType(type) != null)
            {
                if (string.IsNullOrEmpty(source)) return null;
                type = Nullable.GetUnderlyingType(type);
            }

            if ( type == typeof(byte[]) )
                return System.Convert.FromBase64String(source);

            if (type.IsEnum)
                return Enum.Parse(type, source);

            switch ( name )
            {
                case nameof(Boolean): return System.Xml.XmlConvert.ToBoolean(source);
                case nameof(Char): return System.Xml.XmlConvert.ToChar(source);
                case nameof(Decimal): return System.Xml.XmlConvert.ToDecimal(source);
                case nameof(SByte): return System.Xml.XmlConvert.ToSByte(source);
                case nameof(Int16): return System.Xml.XmlConvert.ToInt16(source);
                case nameof(Int32): return System.Xml.XmlConvert.ToInt32(source);
                case nameof(Int64): return System.Xml.XmlConvert.ToInt64(source);
                case nameof(Byte): return System.Xml.XmlConvert.ToByte(source);
                case nameof(UInt16): return System.Xml.XmlConvert.ToUInt16(source);
                case nameof(UInt32): return System.Xml.XmlConvert.ToUInt32(source);
                case nameof(UInt64): return System.Xml.XmlConvert.ToUInt64(source);
                case nameof(Single): return System.Xml.XmlConvert.ToSingle(source);
                case nameof(Double): return System.Xml.XmlConvert.ToDouble(source);
                case nameof(TimeSpan): return System.Xml.XmlConvert.ToTimeSpan(source);
                case nameof(DateTime): return System.Xml.XmlConvert.ToDateTime(source, System.Xml.XmlDateTimeSerializationMode.RoundtripKind);
                case nameof(DateTimeOffset): return System.Xml.XmlConvert.ToDateTimeOffset(source);
                case nameof(Guid): return System.Xml.XmlConvert.ToGuid(source);
                case nameof(Uri): return new Uri(source);
                case nameof(String): return source;
            }

            return global::System.Convert.ChangeType(source, type);
        }

        /// <summary>
        /// Read single object
        /// </summary>
        /// <typeparam name="TModel"></typeparam>
        /// <param name="reader"></param>
        /// <param name="cancellationToken"></param>
        /// <returns></returns>
        public static TModel ReadObject<TModel>(this System.Xml.XmlReader reader, CancellationToken cancellationToken = default(CancellationToken))
             where TModel : class, new()
        {
            // check for end of collection
            if (reader.NodeType != System.Xml.XmlNodeType.Element)
                return null;

            var element = XNode.ReadFrom(reader) as XElement;
            if (element == null) return null;

            var type = typeof(TModel);
            var model = new TModel();

            // parse
            foreach (var node in element.Descendants())
            {
                var property = type.GetProperty(node.Name.LocalName);
                if (property == null) continue;
                if (property.GetSetMethod() == null) continue;
                if (cancellationToken.IsCancellationRequested) return null;

                object value;

                try
                {
                    value = Convert(node.Value, property.PropertyType);
                }
                catch( Exception ex )
                {
                    Debug.WriteLine("Failed to convert property {0} of type {1} from {2}: {3}", property.Name, property.PropertyType.Name, node.Value, ex.Message);
                    continue;
                }

                try
                {
                    property.SetValue(model, value);
                }
                catch( Exception ex )
                {
                    Debug.WriteLine("Failed to set property {0} of type {1} to {2}: {3}", property.Name, property.PropertyType.Name, value, ex.Message);
                }
            }

            return model;
        }

        public static IEnumerable<TModel> ReadCollectionAsync<TModel>( this System.Xml.XmlReader reader, string name, CancellationToken cancellationToken = default(CancellationToken) )
             where TModel : class, new()
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                var model = ReadObject<TModel>(reader, cancellationToken);
                if (model == null) yield break;
                else yield return model;
            }
        }
    }
}
