﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Platform.Net
{
    /// <summary>
    /// Collection of authority parameters
    /// </summary>
    public abstract class NetList<T> : NetValue, IList<T>
    {
        #region Construction & Fields

        protected bool _HasPrefix;
        protected bool _HasSuffix;
        protected List<T> _List = new List<T>();

        public NetList()
        {
        }

        protected NetList( NetStringReader reader )
            : base(reader)
        {
        }

        protected NetList(string source)
            : base(source)
        {
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets if list is prefixed
        /// </summary>
        public bool HasPrefix
        {
            get { Decode(); return _HasPrefix; }
            set { Decode(); _HasPrefix = value; }
        }

        /// <summary>
        /// Gets or sets if list is suffixed
        /// </summary>
        public bool HasSuffix
        {
            get { Decode(); return _HasSuffix; }
            set { Decode(); _HasSuffix = value; }
        }

        /// <summary>
        /// Gets the collection item seperator
        /// </summary>
        protected abstract char Seperator { get; }

        /// <summary>
        /// Gets if this list uses item prefixes, item seperators 
        /// or if this is detected dynamic
        /// </summary>
        protected abstract NetListSeperationType Seperation { get; }

        #endregion

        #region Tools

        /// <summary>
        /// Force seperation type onto this collection
        /// </summary>
        protected void ForceSeperationType()
        {
            // check for prefixed list
            if (Seperation == NetListSeperationType.Prefixed)
            {
                _HasPrefix = _List.Count > 0;
                _HasSuffix = false;
            }

            // check for seperated list
            if (Seperation == NetListSeperationType.Seperated)
            {
                _HasPrefix = false;
                _HasSuffix = false;
            }
        }

        #endregion

        #region Encoding / Decoding

        /// <summary>
        /// Create Parameter
        /// </summary>
        /// <param name="source"></param>
        /// <returns></returns>
        protected abstract T CreateListItem(NetStringReader reader);

        /// <summary>
        /// Overridden to enforce seperation type
        /// </summary>
        /// <param name="writer"></param>
        protected override void OnEncode(NetStringWriter writer)
        {
            // force seperation type
            ForceSeperationType();

            // add prefix
            if (_HasPrefix) writer.Write(Seperator);

            // create result
            for (int index = 0; index < _List.Count; index++)
            {
                if (index > 0) writer.Write(Seperator);
                var value = _List[index];
                if (value != null) writer.Write( value.ToString() );
            }

            // add suffix
            if (_HasSuffix) writer.Write(Seperator);
        }

        /// <summary>
        /// Decode from reader
        /// </summary>
        /// <param name="parser"></param>
        internal protected void Decode(NetStringReader reader)
        {
            // create list with parameters
            _List.Clear();

            // first check for prefix
            _HasPrefix = reader.Read(Seperator);

            // read items
            while (!reader.EndOfString)
            {
                var item = CreateListItem(reader);
                if (item == null) break;
                else _List.Add(item);
            }

            // next check for suffix
            _HasSuffix = reader.Read(Seperator);

            // force seperation type
            ForceSeperationType();
        }

        /// <summary>
        /// Decode from string value
        /// </summary>
        protected void Decode( string source )
        {
            // create list with parameters
            _List.Clear();

            // check for prefix & suffix
            _HasPrefix = source.Length > 0 && source[0] == Seperator;
            _HasSuffix = source.Length > 1 && source[source.Length-1] == Seperator;

            // create parameters
            foreach (var item in source.Trim(Seperator).Split(new char[] { Seperator }).Where( s => !string.IsNullOrEmpty(s)) )
                _List.Add(CreateListItem(new NetStringReader(item)));

            // force seperation type
            ForceSeperationType();
        }

        #endregion

        #region IList Interface

        public int IndexOf(T item)
        {
            Decode();
            return _List.IndexOf(item);
        }

        public void Insert(int index, T item)
        {
            Decode();
            _List.Insert(index, item);
        }

        public void RemoveAt(int index)
        {
            Decode();
            _List.RemoveAt(index);
        }

        public T this[int index]
        {
            get { Decode(); return _List[index]; }
            set { Decode(); _List[index] = value; }
        }

        public void Add(T item)
        {
            Decode();
            _List.Add(item);
        }

        public void AddRange(IEnumerable<T> items)
        {
            Decode();
            _List.AddRange(items);
        }

        public void Clear()
        {
            Decode(); 
            _List.Clear();
        }

        public bool Contains(T item)
        {
            Decode(); 
            return _List.Contains(item);
        }

        public void CopyTo(T[] array, int arrayIndex)
        {
            Decode(); 
            _List.CopyTo(array, arrayIndex);
        }

        public int Count
        {
            get { Decode(); return _List.Count; }
        }

        public bool IsReadOnly
        {
            get { return false; }
        }

        public bool Remove(T item)
        {
            Decode(); 
            return _List.Remove(item);
        }

        public IEnumerator<T> GetEnumerator()
        {
            Decode(); 
            return _List.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            Decode(); 
            return _List.GetEnumerator();
        }

        #endregion
    }
}
