﻿using System.Linq;

using Platform.Extensions;

namespace Platform.Net
{
    /// <summary>
    /// Basic collection of parameters
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public abstract class NetParameters<T> : NetList<T>
        where T : NetParameter
    {
        #region Construction & Fields

        public NetParameters()
        {
        }

        protected NetParameters( NetStringReader reader )
            : base(reader)
        {
        }

        protected NetParameters(string source)
            : base( source )
        {
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets, sets or removes a parameter
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public string this[string name]
        {
            get
            {
                Decode();
                var item = this.FirstOrDefault(n => n.Name.IsEqualTo(name, true));
                if (item == null) return null;
                if (item.Value == null) return NetValue.NoValue;
                return item.Value;
            }
            set
            {
                Decode();
                var item = this.FirstOrDefault(n => n.Name.IsEqualTo(name, true));

                if (value == null)
                {
                    if (item != null) Remove(item);
                    return;
                }

                if (value == NetValue.NoValue) 
                    value = null;

                if (item != null) item.Value = value;
                else Add(CreateParameter(name, value));                    
            }
        }

        /// <summary>
        /// Gets a parameter value or a default
        /// </summary>
        /// <param name="name"></param>
        /// <param name="default"></param>
        /// <returns></returns>
        public string this[string name, string @default]
        {
            get
            {
                Decode();
                var item = this.FirstOrDefault(n => n.Name.IsEqualTo(name, true));
                if (item == null) return @default;
                return item.Value??@default;
            }
        }

        /// <summary>
        /// Gets a parameter value or a default
        /// </summary>
        /// <param name="name"></param>
        /// <param name="default"></param>
        /// <returns></returns>
        public string this[string name, string @default, string @novalue]
        {
            get
            {
                Decode();
                var item = this.FirstOrDefault(n => n.Name.IsEqualTo(name, true));
                if (item == null) return @default;
                return item.Value??@novalue;
            }
        }

        #endregion

        #region Encoding / Decoding

        /// <summary>
        /// Create Parameter
        /// </summary>
        /// <param name="name"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        protected abstract T CreateParameter(string name, string value);

        #endregion

        #region Tools

        /// <summary>
        /// Gets existing parameter
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public T GetParameter( string name )
        {
            Decode();
            return this.FirstOrDefault(n => n.Name.IsEqualTo(name, true));
        }

        /// <summary>
        /// Set existing parameter or create new segment
        /// </summary>
        /// <param name="name"></param>
        /// <param name="value"></param>
        public void SetParameter(string name, string value)
        {
            Decode();
            var item = this.FirstOrDefault(n => n.Name.IsEqualTo(name, true));
            if (item != null) item.Value = value;
            else Add(CreateParameter(name, value));
        }

        public bool Contains(string name)
        {
            name = name.ToLower();
            return this.Any(s => s.Name.ToLower() == name);
        }

        public bool Contains(string name, string value)
        {
            name = name.ToLower();
            return this.Any(s => s.Name.ToLower() == name && s.Value == value);
        }

        public bool GetFlag(string name)
        {
            return Contains(name);
        }

        public void Remove( string name )
        {
            for( var item = this.FirstOrDefault( s => s.Name.ToLower() == name ); item != null; 
                item = this.FirstOrDefault( s => s.Name.ToLower() == name ) ) Remove( item );
        }

        public void SetFlag( string name, bool value )
        {
            // create flag
            if( value && !Contains(name) ) 
                Add( CreateParameter( name, null ) );

            // remove flag
            if( !value && Contains(name) )
                Remove( name );
        }

        #endregion

        #region Comparison

        /// <summary>
        /// All parameter should be in other collection.
        /// Order is not important
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public override bool IsMatch(NetValue value)
        {
            if( value.GetType() != GetType() ) return false;
            var other = value as NetParameters<T>;
            if (Count != other.Count) return false;
            foreach (var item in this)
                if (!other.Any(v => v.IsMatch(item))) return false;
            return true;
        }

        #endregion
    }
}
