﻿using System.Linq;

namespace Platform.Net
{
    public abstract class NetValue
    {
        #region Static Helpers

        // indicates a parameter without value
        public const string NoValue = "\0";

        /// <summary>
        /// Check if string is a token
        /// </summary>
        /// <param name="source"></param>
        /// <returns></returns>
        public static bool IsToken(string source)
        {
            for (int i = 0; i < source.Length; i++)
                if (!char.IsLetterOrDigit(source[i]) && !NetStringReader.TOKEN.Contains(source[i]))
                    return false;
            return true;
        }

        #endregion

        #region Construction & Fields

        private string _Source;

        /// <summary>
        /// Construct new value
        /// </summary>
        protected NetValue()
        {
            Initialize();
        }

        /// <summary>
        /// Construct new value from text
        /// </summary>
        /// <param name="source"></param>
        protected NetValue(string source)
        {
            if (source == null) Initialize();
            else _Source = source;
        }

        /// <summary>
        /// Construct new value from parser
        /// </summary>
        /// <param name="source"></param>
        protected NetValue(NetStringReader parser)
        {
            OnDecode(parser);
        }

        /// <summary>
        /// Initialize value
        /// </summary>
        protected virtual void Initialize()
        {
        }

        #endregion

        #region Encoding / Decoding

        /// <summary>
        /// Gets or sets the source string
        /// </summary>
        protected string Source
        {
            get { return _Source; }
            set { _Source = value; }
        }

        /// <summary>
        /// Decode to properties
        /// </summary>
        protected virtual void OnDecode(NetStringReader reader)
        {
        }

        /// <summary>
        /// Encode from properties
        /// </summary>
        /// <returns></returns>
        protected virtual void OnEncode(NetStringWriter writer)
        {
        }

        /// <summary>
        /// Decode for property access
        /// </summary>
        protected void Decode()
        {
            // allready decoded
            if (_Source == null) return;

            // decode
            OnDecode(new NetStringReader(_Source));

            // decode done
            _Source = null;
        }

        /// <summary>
        /// Encode for string access
        /// </summary>
        public virtual string Encode()
        {
            if (_Source == null)
            {
                var writer = new NetStringWriter();
                OnEncode(writer);
                _Source = writer.ToString();
            }

            return _Source;
        }

        /// <summary>
        /// Convert to string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            // if encoded
            if (_Source != null)
                return _Source;

            // encode and return string
            var writer = new NetStringWriter();
            OnEncode(writer);
            return writer.ToString();
        }

        /// <summary>
        /// Check if two values match
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public virtual bool IsMatch(NetValue value)
        {
            if (value.GetType() != GetType()) return false;
            return ToString() == value.ToString();
        }

        #endregion
    }
}
